﻿/********************************************************
 *  ██████╗  ██████╗████████╗██╗
 * ██╔════╝ ██╔════╝╚══██╔══╝██║
 * ██║  ███╗██║        ██║   ██║
 * ██║   ██║██║        ██║   ██║
 * ╚██████╔╝╚██████╗   ██║   ███████╗
 *  ╚═════╝  ╚═════╝   ╚═╝   ╚══════╝
 * Geophysical Computational Tools & Library (GCTL)
 *
 * Copyright (c) 2022  Yi Zhang (yizhang-geo@zju.edu.cn)
 *
 * GCTL is distributed under a dual licensing scheme. You can redistribute 
 * it and/or modify it under the terms of the GNU Lesser General Public 
 * License as published by the Free Software Foundation, either version 2 
 * of the License, or (at your option) any later version. You should have 
 * received a copy of the GNU Lesser General Public License along with this 
 * program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * If the terms and conditions of the LGPL v.2. would prevent you from using 
 * the GCTL, please consider the option to obtain a commercial license for a 
 * fee. These licenses are offered by the GCTL's original author. As a rule, 
 * licenses are provided "as-is", unlimited in time for a one time fee. Please 
 * send corresponding requests to: yizhang-geo@zju.edu.cn. Please do not forget 
 * to include some description of your company and the realm of its activities. 
 * Also add information on how to contact you by electronic and paper mail.
 ******************************************************/

#include "gm_regular_mesh_3d.h"

gctl::gm_regular_mesh_3d::gm_regular_mesh_3d(){}

gctl::gm_regular_mesh_3d::gm_regular_mesh_3d(std::string in_name, std::string in_info, 
	int xbnum, int ybnum, int zbnum, double xmin, double ymin, double zmin, 
	double xsize, double ysize, double zsize) : regular_mesh_3d::regular_mesh_3d(in_name, in_info, 
	xbnum, ybnum, zbnum, xmin, ymin, zmin, xsize, ysize, zsize){}

gctl::gm_regular_mesh_3d::~gm_regular_mesh_3d(){}

void gctl::gm_regular_mesh_3d::gkernel(matrix<double> &out_kernel, const array<point3dc> &obs, 
	gravitational_field_type_e comp_id, verbose_type_e verbose)
{
	gctl::gkernel(out_kernel, elements, obs, comp_id, verbose);
	return;
}

void gctl::gm_regular_mesh_3d::gobser(array<double> &out_data, std::string data_name, const array<point3dc> &obs, 
	gravitational_field_type_e comp_id, verbose_type_e verbose)
{
	array<double> rho(get_elenum(), 0.0);

	meshdata &data = get_data(data_name);
	mesh_data_value_e vtype = data.valtype_;
	if (vtype == Scalar)
	{
		for (int i = 0; i < data.datval_.size(); i++)
		{
			if (!std::isnan(data.datval_[i]))
			{
				rho[i] = data.datval_[i];
			}
		}
	}
	else
	{
		std::string err_str = "Unsupported data value type. From gctl::gm_regular_mesh_3d::gobser(...)";
		throw runtime_error(err_str);
	}

	gctl::gobser(out_data, elements, obs, rho, comp_id, verbose);
	return;
}

void gctl::gm_regular_mesh_3d::magkernel(matrix<double> &out_kernel, const array<point3dc> &obs, 
	const array<double> &inc_deg, const array<double> &dec_deg, double geo_inc, double geo_dec, 
	magnetic_field_type_e comp_id, verbose_type_e verbose)
{
	mag_ele_.resize(get_elenum());
	array<magblock_para> mp(get_elenum());

	for (int i = 0; i < get_elenum(); ++i)
	{
		mp[i].inclina_deg = inc_deg[i];
		mp[i].inclina_deg = dec_deg[i];

		copy_type_block(mag_ele_.get(i), elements.get(i));
		link_entity_attribute(mag_ele_[i],mp.get(i));
	}

	gctl::magkernel(out_kernel, mag_ele_, obs, geo_inc, geo_dec, comp_id, verbose);
	return;
}

void gctl::gm_regular_mesh_3d::magkernel(matrix<double> &out_kernel, const array<point3dc> &obs, 
    const array<double> &inc_deg, const array<double> &dec_deg, magnetic_field_type_e comp_id, verbose_type_e verbose)
{
	mag_ele_.resize(get_elenum());
	array<magblock_para> mp(get_elenum());

	for (int i = 0; i < get_elenum(); ++i)
	{
		mp[i].inclina_deg = inc_deg[i];
		mp[i].inclina_deg = dec_deg[i];

		copy_type_block(mag_ele_.get(i), elements.get(i));
		link_entity_attribute(mag_ele_[i], mp.get(i));
	}

	gctl::magkernel(out_kernel, mag_ele_, obs, comp_id, verbose);
	return;
}

void gctl::gm_regular_mesh_3d::magobser(array<double> &out_data, std::string data_name, const array<point3dc> &obs, 
    const array<double> &inc_deg, const array<double> &dec_deg, double geo_inc, double geo_dec, 
    magnetic_field_type_e comp_id, verbose_type_e verbose)
{
	array<double> sus(get_elenum(), 0.0);

	meshdata &data = get_data(data_name);
	mesh_data_value_e vtype = data.valtype_;
	if (vtype == Scalar)
	{
		for (int i = 0; i < data.datval_.size(); i++)
		{
			if (!std::isnan(data.datval_[i]))
			{
				sus[i] = data.datval_[i];
			}
		}
	}
	else
	{
		std::string err_str = "Unsupported data value type. From gctl::gm_regular_mesh_3d::gobser(...)";
		throw runtime_error(err_str);
	}

	mag_ele_.resize(get_elenum());
	array<magblock_para> mp(get_elenum());

	for (int i = 0; i < get_elenum(); ++i)
	{
		mp[i].inclina_deg = inc_deg[i];
		mp[i].inclina_deg = dec_deg[i];

		copy_type_block(mag_ele_.get(i), elements.get(i));
		link_entity_attribute(mag_ele_[i], mp.get(i));
	}

	gctl::magobser(out_data, mag_ele_, obs, sus, geo_inc, geo_dec, comp_id, verbose);
	return;
}

void gctl::gm_regular_mesh_3d::magobser(array<double> &out_data, std::string data_name, const array<point3dc> &obs, 
    const array<double> &inc_deg, const array<double> &dec_deg, magnetic_field_type_e comp_id, verbose_type_e verbose)
{
	array<double> sus(get_elenum(), 0.0);

	meshdata &data = get_data(data_name);
	mesh_data_value_e vtype = data.valtype_;
	if (vtype == Scalar)
	{
		for (int i = 0; i < data.datval_.size(); i++)
		{
			if (!std::isnan(data.datval_[i]))
			{
				sus[i] = data.datval_[i];
			}
		}
	}
	else
	{
		std::string err_str = "Unsupported data value type. From gctl::gm_regular_mesh_3d::gobser(...)";
		throw runtime_error(err_str);
	}

	mag_ele_.resize(get_elenum());
	array<magblock_para> mp(get_elenum());

	for (int i = 0; i < get_elenum(); ++i)
	{
		mp[i].inclina_deg = inc_deg[i];
		mp[i].inclina_deg = dec_deg[i];

		copy_type_block(mag_ele_.get(i), elements.get(i));
		link_entity_attribute(mag_ele_[i], mp.get(i));
	}

	gctl::magobser(out_data, mag_ele_, obs, sus, comp_id, verbose);
	return;
}